import logging
import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature
from balder.exceptions import AccessToUnmappedVDeviceException


logger = logging.getLogger(__name__)


class ScenarioAddingVDevice2WithTwoMapped(balder.Scenario):

    class Calculator(balder.Device):
        adds = AddCalculateFeature(ANumberProviderDevice2="NumberTwoDevice")

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberTwoDevice.number.set_number(4)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 7, "result is not as expected"

        self.validate_vdevice_references()

    def validate_vdevice_references(self):
        assert self.Calculator.adds.get_number_of_provider_2() == self.NumberTwoDevice.number.get_number(), \
            f"the currently mapped internal vdevice returns an unexpected value " \
            f"`{self.Calculator.adds.get_number_of_provider_2()}` " \
            f"(expected `{self.NumberTwoDevice.number.get_number()}`)"

        try:
            self.Calculator.adds.get_number_of_provider_1()
            assert False, "provider 1 does return a value -> not allowed, because this VDevice was not mapped"
        except AccessToUnmappedVDeviceException as exc:
            logger.info(f'access to unmapped VDevice throws error message: `{str(exc)}`')
            assert exc.args[0] == "it is not allowed to access the attributes of an unmapped VDevice - " \
                                  "did you forget to map it?"

    @balder.fixture(level='variation')
    def fixture_vdevice_access_variation(self):
        self.validate_vdevice_references()
        yield
        self.validate_vdevice_references()

    @balder.fixture(level='testcase')
    def fixture_vdevice_access_testcase(self):
        self.validate_vdevice_references()
        yield
        self.validate_vdevice_references()



import balder
import logging
from .setup_features import SetupFeatureI, SetupFeatureII
from ..lib.connections import AConnection
from ..lib.utils import FixtureReturn
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupA(balder.Setup):
    """This is a setup of category A (exactly the same as scenario A)"""

    class SetupDevice1(balder.Device):
        s_i = SetupFeatureI()

    @balder.connect(SetupDevice1, over_connection=AConnection)
    class SetupDevice2(balder.Device):
        s_ii = SetupFeatureII()

    @classmethod
    @balder.fixture(level="session")
    def fixture_setup_session(cls, balderglob_fixture_session):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_session == FixtureReturn.BALDERGLOB_SESSION

        yield FixtureReturn.SETUP_SESSION

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_session == FixtureReturn.BALDERGLOB_SESSION
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup_setup(self, balderglob_fixture_setup, fixture_setup_session):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_setup, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert balderglob_fixture_setup == FixtureReturn.BALDERGLOB_SETUP
        assert fixture_setup_session == FixtureReturn.SETUP_SESSION

        yield FixtureReturn.SETUP_SETUP

        assert isinstance(self, SetupA)
        assert balderglob_fixture_setup == FixtureReturn.BALDERGLOB_SETUP
        assert fixture_setup_session == FixtureReturn.SETUP_SESSION
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @classmethod
    @balder.fixture(level="scenario")
    def fixture_setup_scenario(cls, balderglob_fixture_scenario, fixture_setup_setup):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO
        assert fixture_setup_setup == FixtureReturn.SETUP_SETUP

        yield FixtureReturn.SETUP_SCENARIO

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO
        assert fixture_setup_setup == FixtureReturn.SETUP_SETUP
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="variation")
    def fixture_setup_variation(self, balderglob_fixture_variation, fixture_setup_scenario):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert balderglob_fixture_variation == FixtureReturn.BALDERGLOB_VARIATION
        assert fixture_setup_scenario == FixtureReturn.SETUP_SCENARIO

        yield FixtureReturn.SETUP_VARIATION

        assert isinstance(self, SetupA)
        assert balderglob_fixture_variation == FixtureReturn.BALDERGLOB_VARIATION
        assert fixture_setup_scenario == FixtureReturn.SETUP_SCENARIO
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @classmethod
    @balder.fixture(level="testcase")
    def fixture_setup_testcase(cls, balderglob_fixture_testcase, fixture_setup_variation):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE
        assert fixture_setup_variation == FixtureReturn.SETUP_VARIATION

        yield FixtureReturn.SETUP_TESTCASE

        assert issubclass(cls, SetupA)
        assert balderglob_fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE
        assert fixture_setup_variation == FixtureReturn.SETUP_VARIATION
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        cls.SetupDevice1.s_i.do_something()
        cls.SetupDevice2.s_ii.do_something()



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0InheritScenario(Base0EnvtesterClass):
    """
    This testcase executes the reduced envtester environment. Only one setup and one scenario of the basic envtester
    environment are implemented here. In addition to that, this environment provides an additional scenario, which is
    a child class of the other scenario. Generally both scenarios will match, but balder should exclude the parent
    class, because parent classes are not allowed to be executed.

    This test secures that only the child-scenario will be executed here.
    """

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioAChild.fixture_session
            {"cls": "ScenarioAChild", "meth": "fixture_session", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioAChild.fixture_setup
            {"cls": "ScenarioAChild", "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioAChild.fixture_scenario
            {"cls": "ScenarioAChild", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioAChild.fixture_variation
            {"cls": "ScenarioAChild", "meth": "fixture_variation", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: ScenarioAChild.fixture_testcase
            {"cls": "ScenarioAChild", "meth": "fixture_testcase", "part": "construction"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # TESTCASE: ScenarioAChild.test_a_1
            {"cls": "ScenarioAChild", "meth": "test_a_1"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: ScenarioAChild.fixture_testcase
            {"cls": "ScenarioAChild", "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioAChild.fixture_variation
            {"cls": "ScenarioAChild", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_variation
            {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioAChild.fixture_scenario
            {"cls": "ScenarioAChild", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_scenario
            {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioAChild.fixture_setup
            {"cls": "ScenarioAChild", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_setup
            {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
            # FIXTURE-TEARDOWN: ScenarioAChild.fixture_session
            {"cls": "ScenarioAChild", "meth": "fixture_session", "part": "teardown"},
            {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



import balder
import logging
from .setup_features import SetupFeatureIII, SetupFeatureIV
from ..lib.connections import BConnection
from ..lib.utils import FixtureReturn
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupB(balder.Setup):
    """This is a setup of category B (exactly the same as scenario B)"""

    class SetupDevice3(balder.Device):
        s_iii = SetupFeatureIII()

    @balder.connect(SetupDevice3, over_connection=BConnection)
    class SetupDevice4(balder.Device):
        s_iv = SetupFeatureIV()

    @balder.fixture(level="session")
    def fixture_session(self, fixture_session):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_session,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        assert isinstance(self, SetupB)
        assert fixture_session == FixtureReturn.BALDERGLOB_SESSION

        yield FixtureReturn.SETUP_B_SESSION

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_session,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self, fixture_session, fixture_setup):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_setup,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        assert isinstance(self, SetupB)
        assert fixture_session == FixtureReturn.SETUP_B_SESSION
        assert fixture_setup == FixtureReturn.BALDERGLOB_SETUP

        yield FixtureReturn.SETUP_B_SETUP

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_setup,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self, fixture_session, fixture_setup, fixture_scenario):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_scenario,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        assert isinstance(self, SetupB)
        assert fixture_session == FixtureReturn.SETUP_B_SESSION
        assert fixture_setup == FixtureReturn.SETUP_B_SETUP
        assert fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO

        yield FixtureReturn.SETUP_B_SCENARIO

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_scenario,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_variation,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        assert isinstance(self, SetupB)
        assert fixture_session == FixtureReturn.SETUP_B_SESSION
        assert fixture_setup == FixtureReturn.SETUP_B_SETUP
        assert fixture_scenario == FixtureReturn.SETUP_B_SCENARIO
        assert fixture_variation == FixtureReturn.BALDERGLOB_VARIATION

        yield FixtureReturn.SETUP_B_VARIATION

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_variation,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation, fixture_testcase):
        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_testcase,  category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()

        assert isinstance(self, SetupB)
        assert fixture_session == FixtureReturn.SETUP_B_SESSION
        assert fixture_setup == FixtureReturn.SETUP_B_SETUP
        assert fixture_scenario == FixtureReturn.SETUP_B_SCENARIO
        assert fixture_variation == FixtureReturn.SETUP_B_VARIATION
        assert fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE

        yield FixtureReturn.SETUP_B_TESTCASE

        RuntimeObserver.add_entry(__file__, SetupB, SetupB.fixture_testcase,  category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice3.s_iii.do_something()
        self.SetupDevice4.s_iv.do_something()



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_scenario_inheritance_use_parent_connections(balder_working_dir):
    """
    This testcase executes the general CALCULATOR example.

    It has two inherited scenarios, while only the parent scenario defines some connections. The environment uses two
    setups - one without connections (would match with child scenario if no parent would exist) and one with the
    required connections, that are only defined in parent scenario class.

    The test validates, that only the correct setup (with connection) will be executed with the child scenario class.
    This secures that the parent connection will be inherited correctly.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.setup_executors) == 1, \
        "not exactly one setup executor found"
    assert session.executor_tree.setup_executors[0].base_setup_class.__class__.__name__ == \
           "SetupPythonAdd", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors) == 1, \
        "not exactly one scenario executor found"
    assert session.executor_tree.setup_executors[0].scenario_executors[0].base_scenario_class.__class__.__name__ == \
           "ScenarioAddingChild", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors[0].variation_executors) == 2, \
        "not exactly two variation executor found"



from __future__ import annotations
from typing import List, Tuple, Generator, Dict, Union, Type, Callable, TYPE_CHECKING

import inspect
from graphlib import TopologicalSorter
from _balder.executor.testcase_executor import TestcaseExecutor
from _balder.scenario import Scenario
from _balder.setup import Setup
from _balder.executor.basic_executor import BasicExecutor
from _balder.executor.setup_executor import SetupExecutor
from _balder.executor.scenario_executor import ScenarioExecutor
from _balder.executor.variation_executor import VariationExecutor
from _balder.exceptions import LostInExecutorTreeException, FixtureReferenceError, UnclearSetupScopedFixtureReference, \
    UnclearUniqueClassReference

if TYPE_CHECKING:
    from _balder.utils import MethodLiteralType
    from _balder.executor.executor_tree import ExecutorTree


class FixtureManager:
    """
    This class is the global fixture manager. It provides various methods to manage fixtures in the balder system.
    """
    #: the ordering for the execution levels
    EXECUTION_LEVEL_ORDER = ['session', 'setup', 'scenario', 'variation', 'testcase']

    def __init__(self, fixtures: Dict[str, Dict[Union[type, None], List[Tuple[MethodLiteralType, Callable]]]]):

        # The first key is the fixture level, the second key is the namespace in which the fixture is defined (for
        # example the scenario class), which describes the definition-scope. As value a list with tuples is returned.
        #  The first element is the type of the method/function and the second is the callable itself.
        self.fixtures: Dict[str, Dict[Union[type, None], List[Tuple[MethodLiteralType, Callable]]]] = fixtures

        # contains all active fixtures with their namespace, their func_type, their callable, the generator object
        # (otherwise an empty generator, if the fixture is not a generator) and the result according to the fixture's
        # construction code (will be cleaned after it leaves a level)
        self.current_tree_fixtures: \
            Dict[str, List[
                Tuple[Union[Type[ExecutorTree], Type[Setup], Type[Scenario]], str, Callable, Generator, object]]] = {}

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def resolve_type_level(self):
        """
        returns a dictionary that holds the executor class as key and the fixture definition that belongs to this
        executor class as key
        """
        from _balder.executor.executor_tree import ExecutorTree

        return {
            ExecutorTree: 'session',
            SetupExecutor: 'setup',
            ScenarioExecutor: 'scenario',
            VariationExecutor: 'variation',
            TestcaseExecutor: 'testcase'
        }

    @property
    def definition_scope_order(self):
        """
        returns a list with the definition scope objects in the priority order (ExecutorTree stands for global fixtures)
        """
        from _balder.executor.executor_tree import ExecutorTree
        return [ExecutorTree, Setup, Scenario]

    @property
    def all_already_run_fixtures(self) -> List[Callable]:
        """
        returns a list of all fixtures that have already been run
        """
        complete_list_in_order = []
        for cur_level in self.EXECUTION_LEVEL_ORDER:
            if cur_level in self.current_tree_fixtures.keys():
                complete_list_in_order += [
                    cur_fixture for _, _, cur_fixture, _, _ in self.current_tree_fixtures[cur_level]]
        return complete_list_in_order

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def get_all_attribute_values(
            self, branch: Union[ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor, TestcaseExecutor],
            callable_func_namespace: Union[Type[ExecutorTree], Type[Scenario], Type[Setup]], callable_func: Callable,
            func_type: str) -> Dict[str, object]:
        """
        This method tries to fill the unresolved function/method arguments of the given fixture callable. For this it
        searches the return values of all already executed fixtures and supplies the argument values of the
        given ``fixture`` callable in a dictionary.

        It automatically manages `self` / `cls` references for func_type `instancemethod` or `classmethod`. It
        only returns the fixture references and ignores `self` / `cls`

        First the method tries to find the arguments in all fixtures that are in the same namespace. If it does not find
        any references, it will look in the next higher scope. It only uses fixture that has run before!

        :param branch: the current active branch

        :param callable_func_namespace: the namespace of the current fixture

        :param callable_func: the callable with the arguments

        :param func_type: returns the func_type of the fixture - depending on this value the first argument will be
                          ignored, because it has to be `cls` for `classmethod` and `self` for `instancemethod`

        :return: the method returns a dictionary with the attribute name as key and the return value as value
        """
        from _balder.executor.executor_tree import ExecutorTree

        arguments = inspect.getfullargspec(callable_func).args
        result_dict = {}

        if func_type in ["classmethod", "instancemethod"]:
            arguments = arguments[1:]

        self._validate_for_unclear_setup_scoped_fixture_reference(
            callable_func_namespace, callable_func, arguments,
            cur_execution_level=self.resolve_type_level[branch.__class__])
        all_possible_namespaces = [ExecutorTree]
        # determine namespaces (in the correct order)
        setup_type = None
        scenario_type = None
        if isinstance(branch, SetupExecutor):
            setup_type = branch.base_setup_class.__class__
            # normally the scenario is None - only if the current namespace is a scenario we can use it
            if issubclass(callable_func_namespace, Scenario):
                scenario_type = callable_func_namespace
            else:
                scenario_type = None
        elif isinstance(branch, ScenarioExecutor):
            setup_type = branch.parent_executor.base_setup_class.__class__
            scenario_type = branch.base_scenario_class.__class__
        elif isinstance(branch, VariationExecutor):
            setup_type = branch.cur_setup_class.__class__
            scenario_type = branch.cur_scenario_class.__class__
        elif isinstance(branch, TestcaseExecutor):
            setup_type = branch.parent_executor.cur_setup_class.__class__
            scenario_type = branch.parent_executor.cur_scenario_class.__class__

        # add to possible namespaces only if the namespace of the current fixture allows this
        if (issubclass(callable_func_namespace, Setup) or issubclass(callable_func_namespace, Scenario)) \
                and setup_type is not None:
            all_possible_namespaces.append(setup_type)
        if issubclass(callable_func_namespace, Scenario) and scenario_type is not None:
            all_possible_namespaces.append(scenario_type)

        for cur_arg in arguments:
            # go to the most specific fixture, because more specific ones overwrite the more global ones
            for cur_possible_namespace in all_possible_namespaces:
                for cur_level in self.EXECUTION_LEVEL_ORDER:
                    if cur_level in self.current_tree_fixtures.keys():
                        # filter only these fixtures that have the same namespace
                        for cur_fixt_namespace, _, cur_fixt, _, cur_fixt_retval in \
                                self.current_tree_fixtures[cur_level]:
                            if cur_fixt_namespace == cur_possible_namespace:
                                if cur_fixt.__name__ == cur_arg:
                                    result_dict[cur_arg] = cur_fixt_retval
            if cur_arg not in result_dict.keys():
                raise FixtureReferenceError(
                        f"the argument `{cur_arg}` in fixture `{callable_func.__qualname__}` could not be resolved")
        return result_dict

    def _validate_for_unclear_setup_scoped_fixture_reference(
            self, fixture_callable_namespace: Union[Type[ExecutorTree], Type[Scenario], Type[Setup]],
            fixture_callable: Callable, arguments: List[str], cur_execution_level: str):
        """
        This helper method validates a given fixture reference for the unclear setup scoped fixture reference.

        The problem:
        A scenario scoped fixture (defined in a scenario) with the execution level SESSION references a fixture (over
        an argument) with a name that is also used for one or more setup scoped fixtures with execution level SESSION.

        For this constellation it is unclear which referenced setup scoped fixture is the correct one.

        .. note::
            Note that the error will also be thrown if only one setup fixture with the referenced name exists. (This
            would work, but it becomes difficult to understand it. Balder prohibits it.

        .. note::
            The error will be thrown if a scenario scoped fixture with the execution level SESSION references a fixture
            name that exists in one or more setup scoped fixtures with the execution level SESSION.

        :param fixture_callable_namespace: the namespace of the current fixture

        :param fixture_callable: the callable with the arguments

        :param arguments: the execution level of the fixture

        :param cur_execution_level: the current execution level that is currently active
        """
        # this method is only interested in fixtures with execution level SESSION!

        if isinstance(fixture_callable_namespace, Scenario) and cur_execution_level == "session":
            all_setup_scoped_fixtures = []

            for cur_namespace_type, fixtures in self.fixtures.get('session', {}).items():
                if cur_namespace_type is not None and issubclass(cur_namespace_type, Setup):
                    all_setup_scoped_fixtures += [cur_fixt.__name__ for _, cur_fixt in fixtures]
            for cur_arg in arguments:
                if cur_arg in all_setup_scoped_fixtures:
                    raise UnclearSetupScopedFixtureReference(
                        f"your fixture `{fixture_callable.__name__}` has the definition scope SCENARIO and tries to "
                        f"access a fixture name `{cur_arg}` - the environment has one or more SESSION "
                        f"fixtures with the definition scope that have the same name - please rename them!")

    def _sort_fixture_list_of_same_definition_scope(
            self, fixture_namespace_dict: Dict[Tuple[str, Union[Type[ExecutorTree], Type[Scenario], Type[Setup]]],
                                               List[object]], outer_scope_fixtures: List[object]) \
            -> List[Tuple[Union[Type[ExecutorTree], Type[Scenario], Type[Setup]], str, Callable]]:
        """
        This is a helper method that allows to sort the fixtures given in `fixture_namespace_dict`, depending on their
        arguments.
        It also checks that every fixture that is mentioned in the arguments is available and an item of
        ``fixture_namespace_dict`` or ``outer_scope_fixtures``. If the method detects a loop between some fixtures, the
        method throws an exception.

        :param fixture_namespace_dict: the unordered dictionary, where the definition scope namespace object is the key
                                       and a list with the fixture tuple is the value

        :param outer_scope_fixtures: the list of fixtures that have already been ordered as elements before the given
                                     one in ``fixture_list`` (for example the item from another **definition-scope** or
                                     from another **execution-level**)

        :return: the list of tuples where the first element is the definition scope object, the second element is the
                 func_type of the fixture and the third is the fixture callable itself
        """
        if len(fixture_namespace_dict) == 0:
            return []

        # returns the func_type for the given fixture
        fixture_func_types = {}
        for cur_namespace_type, cur_fixture_list in fixture_namespace_dict.items():
            for cur_func_type, cur_fixture in cur_fixture_list:
                fixture_func_types[cur_fixture] = cur_func_type

        # - if we want to access a fixture (over name) from another fixture we have to be careful!
        #   -> if we try to access a fixture with a fixture which is not part of the same namespace scope -> ERROR
        #   -> if we try to access a fixture with a fixture which is part of the same namespace scope -> this fixture
        #      has to run before the other (and return the value to the referencing one)

        # now resolve dependencies, by returning them as single items and add them to the TopologicalSorter
        sorter = TopologicalSorter()

        for cur_namespace_type, cur_fixture_list in fixture_namespace_dict.items():
            for _, cur_fixture in cur_fixture_list:
                # add the item to the sorter, because it should always be executed (also if it is not referenced from
                # another one)
                sorter.add((cur_namespace_type, cur_fixture))
                # determine all function/method arguments that has to be resolved for this fixture
                cur_fixture_args = inspect.getfullargspec(cur_fixture).args
                if fixture_func_types[cur_fixture] in ["instancemethod", "classmethod"]:
                    # this is a class method (remove `cls`) or an instance method (remove `self`)
                    cur_fixture_args = cur_fixture_args[1:]
                # now try to find a possible fixture for every argument
                for cur_arg in cur_fixture_args:
                    # search the fixture in current namespace
                    cur_fixture_list_as_names = [search_fixture.__name__ for _, search_fixture in cur_fixture_list]
                    found = False
                    # found referenced fixture in current namespace and definition scope
                    if cur_arg in cur_fixture_list_as_names:
                        _, cur_arg_fixture = cur_fixture_list[cur_fixture_list_as_names.index(cur_arg)]
                        # only add fixture if the referenced one is not the same as the current fixture
                        if cur_fixture != cur_arg_fixture:
                            sorter.add((cur_namespace_type, cur_fixture), (cur_namespace_type, cur_arg_fixture))
                            found = True
                    # did not find referenced fixture in current namespace and definition scope -> search for it in
                    #  outer scopes
                    if not found:
                        # search it in outer fixture list to secure that they exist
                        if cur_arg not in \
                                [cur_outer_fixture.__name__ for cur_outer_fixture in outer_scope_fixtures]:
                            raise FixtureReferenceError(
                                f"can not find a previous fixture with the name `{cur_arg}`")

        return [(cur_definition_scope, fixture_func_types[cur_func], cur_func)
                for cur_definition_scope, cur_func in sorter.static_order()]

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def is_allowed_to_enter(
            self, branch: Union[BasicExecutor, ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor,
                                TestcaseExecutor]) -> bool:
        """
        This method return true if the given branch can be entered, otherwise false
        """
        execution_level = self.resolve_type_level[branch.__class__]
        return execution_level not in self.current_tree_fixtures.keys()

    def is_allowed_to_leave(
            self, branch: Union[BasicExecutor, ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor,
                                TestcaseExecutor]) \
            -> bool:
        """
        This method returns true if the given branch can be left now (there exist entries from earlier run enter()
        for this branch), otherwise false
        """
        execution_level = self.resolve_type_level[branch.__class__]
        return execution_level in self.current_tree_fixtures.keys()

    def enter(self, branch: Union[BasicExecutor, ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor,
                                  TestcaseExecutor]):
        """
        With this method you enter a branch for the fixture manager in order to execute the fixtures contained in it

        :param branch: specifies the element of the ExecutorTree that should be entered (note that the current position
                       is very important here)

        :raise BalderFixtureException: is thrown if an error occurs while executing a user fixture
        """
        execution_level = self.resolve_type_level[branch.__class__]

        if not self.is_allowed_to_enter(branch):
            raise LostInExecutorTreeException(
                "the current branch that should be entered is not allowed, because other branches weren't left yet")

        def empty():
            yield None
        # now iterate over all fixtures that should be executed in this enter() call
        #  -> collect them with all different DEFINITION-SCOPES
        for cur_definition_scope in self.definition_scope_order:
            cur_fixture_list = self.get_all_fixtures_for_current_level(branch=branch).get(cur_definition_scope)
            for cur_scope_namespace_type, cur_fixture_func_type, cur_fixture in cur_fixture_list:
                try:
                    if cur_fixture_func_type in ["function", "staticmethod"]:
                        # fixture is a function or a staticmethod - no first special attribute
                        kwargs = self.get_all_attribute_values(branch, cur_scope_namespace_type, cur_fixture,
                                                               cur_fixture_func_type)
                        cur_generator = cur_fixture(**kwargs)
                    elif cur_fixture_func_type == "classmethod":
                        kwargs = self.get_all_attribute_values(branch, cur_scope_namespace_type, cur_fixture,
                                                               cur_fixture_func_type)
                        cur_generator = cur_fixture(cur_scope_namespace_type, **kwargs)
                    elif cur_fixture_func_type == "instancemethod":
                        self_reference = branch.get_all_base_instances_of_this_branch(
                            with_type=cur_scope_namespace_type, only_runnable_elements=True)
                        if len(self_reference) != 1:
                            raise UnclearUniqueClassReference(
                                f"can not find exactly one reference of the class "
                                f"`{cur_scope_namespace_type.__name__}` in current tree branch")
                        kwargs = self.get_all_attribute_values(branch, cur_scope_namespace_type, cur_fixture,
                                                               cur_fixture_func_type)
                        cur_generator = cur_fixture(self_reference[0], **kwargs)
                    else:
                        raise ValueError(f"found illegal value for func_type `{cur_fixture_func_type}` for fixture "
                                         f"`{cur_fixture.__name__}`")
                    if isinstance(cur_generator, Generator):
                        cur_retvalue = next(cur_generator)
                    else:
                        cur_retvalue = cur_generator
                        cur_generator = empty()
                        next(cur_generator)
                    # add the executed fixtures to global reference
                    if execution_level not in self.current_tree_fixtures.keys():
                        self.current_tree_fixtures[execution_level] = []
                    self.current_tree_fixtures[execution_level].append(
                        (cur_scope_namespace_type, cur_fixture_func_type, cur_fixture, cur_generator, cur_retvalue))
                except StopIteration:
                    pass
                # every other exception that is thrown, will be recognized and rethrown

    def leave(self, branch: Union[BasicExecutor, ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor,
                                  TestcaseExecutor]):
        """
        With this method you leave a previously entered branch and execute the teardown code of the fixtures. Note that
        you can only leave the branch that you entered before!

        :param branch: specifies the element of the ExecutorTree that should be left (note that the current position
                       is very important here)
        """
        execution_level = self.resolve_type_level[branch.__class__]
        if execution_level not in self.current_tree_fixtures.keys():
            raise LostInExecutorTreeException("can not leave the current branch, because it was not entered before")

        current_tree_fixtures_reversed = self.current_tree_fixtures[execution_level]
        current_tree_fixtures_reversed.reverse()
        exception = None
        for _, _, _, cur_generator, _ in current_tree_fixtures_reversed:
            try:
                next(cur_generator)
            except StopIteration:
                pass
            except Exception as exc:
                if not exception:
                    # only save the first exception
                    exception = exc

        # reset the left location
        del self.current_tree_fixtures[execution_level]

        if exception:
            raise exception

    def get_all_fixtures_for_current_level(
            self, branch: Union[ExecutorTree, SetupExecutor, ScenarioExecutor, VariationExecutor, TestcaseExecutor]) \
            -> Dict[Union[Type[ExecutorTree], Type[Scenario], Type[Setup]],
                    List[Tuple[Union[Type[ExecutorTree], Type[Scenario], Type[Setup]], str, object]]]:
        """
        This method delivers all fixtures which should be executed for the given branch of the executor tree.

        The method collects the following fixtures from all three possible DEFINITION-SCOPES:
        1. collect them from balderglob.py (see ``FixtureManager._glob_fixtures``)
        2. collect them from current used Setup OR if no specific one is set (because we are in a higher branch):
           collect them from all that are part of the current branch
        3. collect them from current used Scenario OR if no specific one is set (because we are in a higher branch):
           collect them from all that are part of the current branch

        :param branch: the branch which is currently active

        :return: a dictionary where the definition object (DEFINITION SCOPE) is the key and a list is the
                 value - this value list has tuple elements as children, where the namespace class (:class:`Scenario`
                 or :class:`Setup`) as first argument, the fixture func_type as second and the fixture callable as third
                 argument (this list is ordered after the call hierarchy)
        """
        from _balder.executor.executor_tree import ExecutorTree
        # current relevant EXECUTION LEVEL - all other levels are not relevant for this call
        cur_execution_level = self.resolve_type_level[branch.__class__]
        # get all fixtures of the current relevant level
        fixtures_of_exec_level = self.fixtures.get(cur_execution_level, {})

        all_fixtures = {}
        # get all relevant fixtures of `balderglob.py` (None is key for balderglob fixtures)
        glob_fixtures = fixtures_of_exec_level.get(None, [])
        all_fixtures[ExecutorTree] = {}
        all_fixtures[ExecutorTree][ExecutorTree] = glob_fixtures
        # get all relevant fixtures with definition scope "setup"
        all_fixtures[Setup] = {}
        for cur_setup in branch.get_all_base_instances_of_this_branch(Setup, only_runnable_elements=True):
            # check if there exists fixtures for the current setup
            cur_setup_fixtures = fixtures_of_exec_level.get(cur_setup.__class__, [])
            if cur_setup_fixtures:
                all_fixtures[Setup][cur_setup.__class__] = cur_setup_fixtures

        # get all relevant fixtures with definition scope "scenario"
        all_fixtures[Scenario] = {}
        for cur_scenario in branch.get_all_base_instances_of_this_branch(Scenario, only_runnable_elements=True):
            cur_scenario_fixtures = fixtures_of_exec_level.get(cur_scenario.__class__, [])
            if cur_scenario_fixtures:
                all_fixtures[Scenario][cur_scenario.__class__] = cur_scenario_fixtures

        ordered_fixtures = {}
        # Now the basic order is: [All of ExecutorTree] -> [All of Setup] -> [All of Scenario]
        #  but the order within these DEFINITION SCOPES has to be determined now!
        outer_scope_fixtures = self.all_already_run_fixtures
        ordered_fixtures[ExecutorTree] = self._sort_fixture_list_of_same_definition_scope(
            fixture_namespace_dict=all_fixtures[ExecutorTree], outer_scope_fixtures=outer_scope_fixtures)

        outer_scope_fixtures = \
            self.all_already_run_fixtures + \
            [cur_fixture for _, _, cur_fixture in ordered_fixtures[ExecutorTree]]
        ordered_fixtures[Setup] = self._sort_fixture_list_of_same_definition_scope(
            fixture_namespace_dict=all_fixtures[Setup], outer_scope_fixtures=outer_scope_fixtures)

        outer_scope_fixtures = \
            self.all_already_run_fixtures + \
            [cur_fixture for _, _, cur_fixture in ordered_fixtures[ExecutorTree]] + \
            [cur_fixture for _, _, cur_fixture in ordered_fixtures[Setup]]
        ordered_fixtures[Scenario] = self._sort_fixture_list_of_same_definition_scope(
            fixture_namespace_dict=all_fixtures[Scenario], outer_scope_fixtures=outer_scope_fixtures)

        return ordered_fixtures



from abc import ABC
from .base_0_envtester_class import Base0EnvtesterClass


class Base01EnvtesterMethvarClass(Base0EnvtesterClass, ABC):
    """
    This is a base class for all tests that want to use the 01_envtester_methvar environment.
    """
    pass



import balder
from ..balderglob import RuntimeObserver


class IllegalFeature(balder.Feature):
    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, IllegalFeature, IllegalFeature.do_something,
            "enter `IllegalFeature.do_something`", category="feature")


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")


class FeatureII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")


class FeatureOfRelevanceLvl1(balder.Feature):
    """this feature will be used in scenario"""

    class VDeviceWithI(balder.VDevice):
        i = FeatureI()

    class VDeviceWithII(balder.VDevice):
        i = FeatureII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureOfRelevanceLvl1, FeatureOfRelevanceLvl1.do_something,
            "enter `FeatureOfRelevanceLvl1.do_something`", category="feature")




import balder
from ..balderglob import RuntimeObserver


class FeatureVDeviceI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVDeviceI, FeatureVDeviceI.do_something, "enter `FeatureVDeviceI.do_something`",
            category="feature")


class FeatureVDeviceII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVDeviceII, FeatureVDeviceII.do_something, "enter `FeatureVDeviceII.do_something`",
            category="feature")


class FeatureVDeviceIII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureVDeviceIII, FeatureVDeviceIII.do_something, "enter `FeatureVDeviceIII.do_something`",
            category="feature")


@balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
class BaseMethVarFeature(balder.Feature):

    class VDeviceI(balder.VDevice):
        feat = FeatureVDeviceI()

    class VDeviceII(balder.VDevice):
        feat = FeatureVDeviceII()

    class VDeviceIII(balder.VDevice):
        feat = FeatureVDeviceIII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, BaseMethVarFeature, BaseMethVarFeature.do_something, "enter `BaseMethVarFeature.do_something`",
            category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
    def do_something_as_var(self):
        RuntimeObserver.add_entry(
            __file__, BaseMethVarFeature, BaseMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI><BaseMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
    def do_something_as_var(self):
        RuntimeObserver.add_entry(
            __file__, BaseMethVarFeature, BaseMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceII><BaseMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
    def do_something_as_var(self):
        RuntimeObserver.add_entry(
            __file__, BaseMethVarFeature, BaseMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceIII><BaseMethVarFeature.do_something_as_var>", category="feature")


@balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
class ScenarioMethVarFeature(BaseMethVarFeature):
    """This is our direct Scenario feature."""
    class VDeviceI(BaseMethVarFeature.VDeviceI):
        pass

    class VDeviceII(BaseMethVarFeature.VDeviceII):
        pass

    class VDeviceIII(BaseMethVarFeature.VDeviceIII):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, ScenarioMethVarFeature, ScenarioMethVarFeature.do_something,
            "enter `ScenarioMethVarFeature.do_something`",
            category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, ScenarioMethVarFeature, ScenarioMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI><ScenarioMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, ScenarioMethVarFeature, ScenarioMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceII><ScenarioMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, ScenarioMethVarFeature, ScenarioMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceIII><ScenarioMethVarFeature.do_something_as_var>", category="feature")



from __future__ import annotations
from typing import Type, Union

import re
from _balder.device import Device
from _balder.connection import Connection
from _balder.controllers import DeviceController


def connect(with_device: Union[Type[Device], str], over_connection: Union[Type[Connection], Connection],
            self_node_name: str = None, dest_node_name: str = None):
    """
    This decorator connects two devices with each other. It can be used for scenarios as well as setup devices.

    :param with_device: that's the :class:`Device` that should be connected to the decorated device

    :param over_connection: that's the connection tree that should connect the devices with each other

    :param self_node_name: the node name of this device (if this param is not given, balder automatically generates a
                           new node name in format `n{unique device counter}`)

    :param dest_node_name: the node name of the destination device, given with `with_device` (if this param is not
                           given, balder automatically generates a new node name in format `n{unique device counter}`)
    """
    allowed_regex_auto_node_names = r"n[0-9]+"

    if not isinstance(with_device, str) and not issubclass(with_device, Device):
        raise ValueError("the value of `with_device` must be a `Device` (or a subclass thereof) or the device name "
                         "as a string")
    if isinstance(over_connection, tuple):
        # doesn't make sense, because we can describe what's needed in one scenario with several `@connect` decorations
        # anyway. We are describing a setup, therefore an AND link makes no sense here either
        raise TypeError("an AND link (signaled via the tuple) of the connection is not possible here - for further "
                        "separate connections use the `@connect` decorator again")
    if isinstance(over_connection, type):
        if not issubclass(over_connection, Connection):
            raise TypeError("the type of `over_connection` must be a `Connection` (or a subclass of it)")
    elif not isinstance(over_connection, Connection):
        raise TypeError("the type of `over_connection` must be a `Connection` (or a subclass of it)")

    if self_node_name is not None:
        if not isinstance(self_node_name, str):
            raise TypeError("the type of `node_name` must be a `str`")
        if re.match(allowed_regex_auto_node_names, self_node_name):
            raise ValueError(
                f"the given `self_node_name` matches the regular expression `{allowed_regex_auto_node_names}` that is "
                f"reserved for internal node naming and should not be used by you")
    if dest_node_name is not None:
        if not isinstance(dest_node_name, str):
            raise TypeError("the type of `node_name` must be a `str`")
        if re.match(allowed_regex_auto_node_names, dest_node_name):
            raise ValueError(
                f"the given `dest_node_name` matches the regular expression `{allowed_regex_auto_node_names}` that is "
                f"reserved for internal node naming and should not be used by you")

    class MyDecorator:
        """
        This decorator will add the connection to the device that is decorated. If the `with_device` is no other device
        class (given as string) this decorator can not handle the node allocation and the device resolving (because not
        all devices are resolved yet). In this case the function adds the device-string in the :class:`Connection`
        object and if the `dest_node_name` is None, it will also set this value to None. This secure that the
        collector will create a unique auto node for it.
        """
        def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument

            decorated_cls = args[0]
            nonlocal with_device
            nonlocal over_connection
            nonlocal self_node_name
            nonlocal dest_node_name

            this_outer_class_name, _ = decorated_cls.__qualname__.split('.')
            if not isinstance(with_device, str):
                other_outer_class_name, _ = with_device.__qualname__.split('.')
                if this_outer_class_name != other_outer_class_name:
                    raise ValueError(
                        f"the given device is not mentioned in this setup/scenario - please create a new "
                        f"direct inner device class, it can be inherited from `{with_device.__qualname__}`")
            decorated_cls_device_controller = DeviceController.get_for(decorated_cls)

            # if required give auto name to nodes
            if self_node_name is None:
                self_node_name = decorated_cls_device_controller.get_new_empty_auto_node()
            if dest_node_name is None and not isinstance(with_device, str):
                with_device_controller = DeviceController.get_for(with_device)
                dest_node_name = with_device_controller.get_new_empty_auto_node()

            cur_cnn_instance = None
            if isinstance(over_connection, Connection):
                # already instantiated because it comes back from a `based_on` - clone it to secure that it was not used
                #  in another `@connect()` decorator (and also remove possible metadata from the new clone)
                cur_cnn_instance = over_connection.clone()
                cur_cnn_instance.set_metadata_for_all_subitems(None)
            elif issubclass(over_connection, Connection):
                # not instantiated -> instantiate it
                cur_cnn_instance = over_connection()
            cur_cnn_instance.set_devices(from_device=decorated_cls, to_device=with_device)
            cur_cnn_instance.update_node_names(from_device_node_name=self_node_name, to_device_node_name=dest_node_name)

            decorated_cls_device_controller.add_new_raw_connection(cur_cnn_instance)
            return decorated_cls
    return MyDecorator



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_01_envtester_methvar_class import Base01EnvtesterMethvarClass


class Test01DiffWithVdeviceAndCnns(Base01EnvtesterMethvarClass):
    """
    This testcase executes the basic example and checks if the tree ends with the result SUCCESS.

    It uses a modified version of the basic environment for method-variation testing. The main feature that is used for
    the test, is the `BaseMethVarFeature`. It still uses the inherit-chain, but in this example with three different
    VDevices. The last setup layer feature (`SetupMethVarFeature`) defines different method variations for all the
    three VDevices. It uses single connection, AND- and OR-combinations of them.
    The environment provides some different setups, that use different VDevice mappings (on setup level) and
    have different connection between the devices (on setup level). These setups provide some different examples, like
    some direct mappings, but also a lot of inner-sub-tree mappings (contained-in).

    The test secures that the correct method variations will be called and the test session terminated with SUCCESS.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},

            [
                (
                    # FIXTURE-CONSTRUCTION: SetupIA.fixture_session
                    {"cls": "SetupIA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupIB.fixture_session
                    {"cls": "SetupIB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupIIA.fixture_session
                    {"cls": "SetupIIA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupIIAOrB.fixture_session
                    {"cls": "SetupIIAOrB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupII.fixture_session
                    {"cls": "SetupIIB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupII.fixture_session
                    {"cls": "SetupIIIAAndB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupII.fixture_session
                    {"cls": "SetupIIIAOrB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupIII.fixture_session
                    {"cls": "SetupIIIB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                )
            ],

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            [
                self._generate_expected_data_from_setup_on("SetupIA", "VDeviceI", "VDeviceI", "VDeviceI",
                                                           "VDeviceI|AConnection"),
                self._generate_expected_data_from_setup_on("SetupIB", "VDeviceI", "VDeviceI", "VDeviceI",
                                                           "VDeviceI|BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIA", "VDeviceII", "VDeviceII", "VDeviceII",
                                                           "VDeviceII|AConnection+OR+BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIAOrB", "VDeviceII", "VDeviceII", "VDeviceII",
                                                           "VDeviceII|AConnection+OR+BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIB", "VDeviceII", "VDeviceII", "VDeviceII",
                                                           "VDeviceII|AConnection+OR+BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIIAAndB", "VDeviceIII", "VDeviceIII", "VDeviceIII",
                                                           "VDeviceIII|AConnection+AND+BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIIAOrB", "VDeviceIII", "VDeviceIII", "VDeviceIII",
                                                           "VDeviceIII|BConnection"),
                self._generate_expected_data_from_setup_on("SetupIIIB", "VDeviceIII", "VDeviceIII", "VDeviceIII",
                                                           "VDeviceIII|BConnection"),
            ],

            # FIXTURE-TEARDOWN: ScenarioA.fixture_session
            {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            [
                (
                    # FIXTURE-TEARDOWN: SetupIA.fixture_session
                    {"cls": "SetupIA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIB.fixture_session
                    {"cls": "SetupIB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIA.fixture_session
                    {"cls": "SetupIIA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIAOrB.fixture_session
                    {"cls": "SetupIIAOrB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIB.fixture_session
                    {"cls": "SetupIIB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIIAAndB.fixture_session
                    {"cls": "SetupIIIAAndB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIIAOrB.fixture_session
                    {"cls": "SetupIIIAOrB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupIIIB.fixture_session
                    {"cls": "SetupIIIB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
                )
            ],

            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere)
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS

    def _generate_expected_data_from_setup_on(
            self, setup_class_name, base_feature_identifier, scenario_feature_identifier,
            between_feature_identifier, setup_feature_identifier):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_setup
            {"cls": setup_class_name, "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_scenario
            {"cls": setup_class_name, "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_variation
            {"cls": setup_class_name, "meth": "fixture_variation", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},

            # FIXTURE-CONSTRUCTION: SetupI.fixture_testcase
            {"cls": setup_class_name, "meth": "fixture_testcase", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # TESTCASE: ScenarioA.test_a_1
            {"cls": "ScenarioA", "meth": "test_a_1"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: SetupI.fixture_testcase
            {"cls": setup_class_name, "meth": "fixture_testcase", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
            {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: SetupI.fixture_variation
            {"cls": setup_class_name, "meth": "fixture_variation", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},
            {"cls": "BaseMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{base_feature_identifier}><BaseMethVarFeature.do_something_as_var>"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{scenario_feature_identifier}><ScenarioMethVarFeature.do_something_as_var>"},
            {"cls": "BetweenMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{between_feature_identifier}><BetweenMethVarFeature.do_something_as_var>"},
            {"cls": "SetupMethVarFeature", "meth": "do_something_as_var", "category": "feature",
             "msg": f"MethodVariation<{setup_feature_identifier}><SetupMethVarFeature.do_something_as_var>"},

            # FIXTURE-TEARDOWN: balderglob_fixture_variation
            {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
            {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: SetupI.fixture_scenario
            {"cls": setup_class_name, "meth": "fixture_scenario", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: balderglob_fixture_scenario
            {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},

            # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
            {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "ScenarioMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: SetupI.fixture_setup
            {"cls": setup_class_name, "meth": "fixture_setup", "part": "teardown"},
            {"cls": "FeatureVDeviceI", "meth": "do_something", "category": "feature"},
            {"cls": "SetupMethVarFeature", "meth": "do_something", "category": "feature"},

            # FIXTURE-TEARDOWN: balderglob_fixture_setup
            {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
        )



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_setup_inheritance_use_parent_connections(balder_working_dir):
    """
    This testcase executes the general CALCULATOR example.

    It has two inherited setups, while only the parent setup defines some connections. The environment defines one
    scenario, that requires the connections, that are provided in the parent setup.

    The test validates, that only the setup will match with the child and both variations are executed as expected.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.setup_executors) == 1, \
        "not exactly one setup executor found"
    assert session.executor_tree.setup_executors[0].base_setup_class.__class__.__name__ == \
           "SetupPythonAddChild", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors) == 1, \
        "not exactly one scenario executor found"
    assert session.executor_tree.setup_executors[0].scenario_executors[0].base_scenario_class.__class__.__name__ == \
           "ScenarioAdding", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors[0].variation_executors) == 2, \
        "not exactly two variation executor found"



from __future__ import annotations
from typing import Type, Dict, Union

import logging
import inspect
from _balder.vdevice import VDevice
from _balder.feature import Feature
from _balder.controllers.base_device_controller import BaseDeviceController
from _balder.exceptions import DeviceScopeError, VDeviceResolvingError

logger = logging.getLogger(__file__)


class VDeviceController(BaseDeviceController):
    """
    This is the controller class for :class:`VDevice` items.
    """
    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing VDevices and its corresponding controller object
    _items: Dict[Type[VDevice], VDeviceController] = {}

    def __init__(self, related_cls, _priv_instantiate_key):
        super().__init__()

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != VDeviceController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`VDeviceController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, VDevice):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{VDevice.__name__}`')
        if related_cls == VDevice:
            raise TypeError(f'the attribute `related_cls` is `{VDevice.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[VDevice]) -> VDeviceController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if VDeviceController._items.get(related_cls) is None:
            item = VDeviceController(related_cls, _priv_instantiate_key=VDeviceController.__priv_instantiate_key)
            VDeviceController._items[related_cls] = item

        return VDeviceController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[VDevice]:
        return self._related_cls

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_outer_class(self) -> Union[Type[Feature], None]:
        """
        This method delivers the outer class of this device. In Balder, this has to be a :class:`Feature`.
        """

        if self.related_cls.__qualname__.count('.') == 0:
            raise DeviceScopeError("the current device class is no inner class")
        all_splits = self.related_cls.__qualname__.split('.')

        outer_class = [cur_class for cur_name, cur_class in inspect.getmembers(
            inspect.getmodule(self.related_cls), inspect.isclass) if cur_name == all_splits[0]][0]
        # only get the next outer class (not higher)
        for cur_class_name in all_splits[1:-1]:
            outer_class = getattr(outer_class, cur_class_name)
        # instantiate it to get the real type (not the decorator class of `@for_vdevice`)
        #  we can instantiate it, because we do nothing in it
        outer_class = outer_class().__class__

        all_inner_classes = [cur_inner_class for _, cur_inner_class in inspect.getmembers(outer_class, inspect.isclass)]
        if self.related_cls in all_inner_classes:
            if not issubclass(outer_class, Feature):
                raise TypeError(f"the found outer class is of type `{outer_class.__name__}` - this is a type which is "
                                f"not allowed (outer class has to be a `Feature` class)")
            return outer_class
        raise RuntimeError(f"can not find the outer class of this vDevice `{self.related_cls.__qualname__}`")

    def get_next_parent_vdevice(self) -> Union[Type[VDevice], None]:
        """
        This method returns the next parent VDevice class, which is still a subclass of :class:`VDevice`. If the next
        parent class is :class:`VDevice`, None will be returned.

        :return: the parent VDevice class or None if no parent exists
        """
        possible_vdevices_of_interest = []
        for cur_vdevice_of_interest in self.related_cls.__bases__:
            if issubclass(cur_vdevice_of_interest, VDevice) and cur_vdevice_of_interest != VDevice:
                possible_vdevices_of_interest.append(cur_vdevice_of_interest)

        if len(possible_vdevices_of_interest) > 1:
            raise VDeviceResolvingError(
                f"the vdevice `{self.related_cls.__name__}` has more than one parent classes from type "
                f"`VDevice` - this is not allowed")

        if len(possible_vdevices_of_interest) == 1:
            # we have found one parent vDevice that has the same name as the cur_vdevice
            return possible_vdevices_of_interest[0]

        # we have no parent vDevice -> there are no parent ConnectionTree
        return None



import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature


class ScenarioAdding(balder.Scenario):

    class Calculator(balder.Device):
        # we could add vDevices here later too -> f.e. `NumberOne="NumberOneDevice", NumberTwo="NumberTwoDevice"`
        adds = AddCalculateFeature()

        @classmethod
        def calculate_add(cls):
            # this should be validated -> on Scenario Level the used methods are not implemented
            # NOTE: Note that only class and static methods work here! We use no device instance and have no access
            cls.adds.get_numbers()
            return cls.adds.add_numbers()

        @staticmethod
        def static_calculate_add():
            # this should be validated -> on Scenario Level the used methods are not implemented
            # NOTE: Note that only class and static methods work here! We use no device instance and have no access
            ScenarioAdding.Calculator.adds.get_numbers()
            return ScenarioAdding.Calculator.adds.add_numbers()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberTwoDevice.number.set_number(4)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        result = self.Calculator.calculate_add()
        assert result == 7, "result with class method is not as expected"

        result = self.Calculator.static_calculate_add()
        assert result == 7, "result with static method is not as expected"



import pathlib
import pytest
import logging

logging.basicConfig(
    format="%(asctime)s [%(threadName)-12.12s] [%(levelname)-7.7s]  %(message)s", level=logging.DEBUG)


@pytest.fixture(scope="function")
def balder_working_dir(request) -> pathlib.Path:
    DIR_NAME = "env"

    filepath = pathlib.Path(request.node.fspath)
    session_working_dir = filepath.parent.joinpath(DIR_NAME)
    if session_working_dir.is_dir():
        return session_working_dir
    else:
        raise NotADirectoryError(f"can not find the testfile related working directory `{DIR_NAME}` for test file "
                                 f"`{filepath.relative_to(request.config.rootdir)}`")



import balder
from ..lib.utils import SharedObj
from .features_setup import PyAddCalculate, PyAddProvideANumber
from .setup_pythonadd import SetupPythonAdd


class SetupPythonAddChild(SetupPythonAdd):

    class Calculator(SetupPythonAdd.Calculator):
        calc = PyAddCalculate()

    class NumberProvider1(SetupPythonAdd.NumberProvider1):
        n = PyAddProvideANumber()

    class NumberProvider2(SetupPythonAdd.NumberProvider2):
        n = PyAddProvideANumber()

    @balder.fixture(level="testcase")
    def cleanup_memory(self):
        SharedObj.shared_mem_list = []



from __future__ import annotations

from _balder.device import Device


class ThisDevice(Device):
    """
    This class describes the test computer on which this code is currently being executed. It represents an abstract
    class that must not be used directly. Rather, it serves as the large superclass over all Balder objects.

    Note that this is the parent class for the :class:`Setup` inner class :class:`Setup.DeviceThis`.
    """



from _balder.exceptions import BalderException, BalderWarning
from _balder.exceptions import FixtureScopeError, FixtureReferenceError, UnclearSetupScopedFixtureReference, \
    UnclearUniqueClassReference, LostInExecutorTreeException, DeviceResolvingException, NodeNotExistsError, \
    DuplicateForVDeviceError, DuplicateBalderSettingError, DeviceOverwritingError, VDeviceOverwritingError, \
    AccessToUnmappedVDeviceException, FeatureOverwritingError, UnknownVDeviceException, RoutingBrokenChainError, \
    IllegalConnectionTypeError, ConnectionMetadataConflictError, DeviceScopeError, ConnectionIntersectionError, \
    UnclearAssignableFeatureConnectionError, InheritanceError, MultiInheritanceError, InnerFeatureResolvingError, \
    VDeviceResolvingError, IllegalVDeviceMappingError, NotApplicableVariationError, UnclearMethodVariationError, \
    UnexpectedPluginMethodReturnValue

__all__ = [

    "BalderException",
    "BalderWarning",
    "FixtureScopeError",
    "FixtureReferenceError",
    "UnclearSetupScopedFixtureReference",
    "UnclearUniqueClassReference",
    "LostInExecutorTreeException",
    "DeviceResolvingException",
    "NodeNotExistsError",
    "DuplicateForVDeviceError",
    "DuplicateBalderSettingError",
    "DeviceOverwritingError",
    "VDeviceOverwritingError",
    "AccessToUnmappedVDeviceException",
    "FeatureOverwritingError",
    "UnknownVDeviceException",
    "RoutingBrokenChainError",
    "IllegalConnectionTypeError",
    "ConnectionMetadataConflictError",
    "DeviceScopeError",
    "ConnectionIntersectionError",
    "UnclearAssignableFeatureConnectionError",
    "InheritanceError",
    "MultiInheritanceError",
    "InnerFeatureResolvingError",
    "VDeviceResolvingError",
    "IllegalVDeviceMappingError",
    "NotApplicableVariationError",
    "UnclearMethodVariationError",
    "UnexpectedPluginMethodReturnValue",
]



from __future__ import annotations
from typing import Type, Dict, List, Tuple, Union

import logging
import inspect
from _balder.device import Device
from _balder.scenario import Scenario
from _balder.connection import Connection
from _balder.controllers.feature_controller import FeatureController
from _balder.controllers.device_controller import DeviceController
from _balder.controllers.normal_scenario_setup_controller import NormalScenarioSetupController
from _balder.exceptions import UnclearAssignableFeatureConnectionError, ConnectionIntersectionError, \
    MultiInheritanceError

logger = logging.getLogger(__file__)


class ScenarioController(NormalScenarioSetupController):
    """
    This is the controller class for :class:`Scenario` items.
    """

    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing scenarios and its corresponding controller object
    _items: Dict[Type[Scenario], ScenarioController] = {}

    def __init__(self, related_cls, _priv_instantiate_key):

        # describes if the current controller is for setups or for scenarios (has to be set in child controller)
        self._related_type = Scenario

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != ScenarioController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`ScenarioController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, Scenario):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Scenario.__name__}`')
        if related_cls == Scenario:
            raise TypeError(f'the attribute `related_cls` is `{Scenario.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[Scenario]) -> ScenarioController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if ScenarioController._items.get(related_cls) is None:
            item = ScenarioController(related_cls, _priv_instantiate_key=ScenarioController.__priv_instantiate_key)
            ScenarioController._items[related_cls] = item

        return ScenarioController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[Scenario]:
        return self._related_cls

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_next_parent_class(self) -> Union[Type[Scenario], None]:
        """
        This method returns the next parent class which is a subclass of the :class:`Scenario` itself.

        :return: returns the next parent class or None if the next parent class is :class:`Scenario`
                 itself
        """
        next_base_class = None
        for cur_base in self.related_cls.__bases__:
            if issubclass(cur_base, Scenario):
                if next_base_class is not None:
                    raise MultiInheritanceError(
                        f"found more than one Scenario parent classes for `{self.related_cls.__name__}` "
                        f"- multi inheritance is not allowed for Scenario/Setup classes")
                next_base_class = cur_base
        if next_base_class == Scenario:
            return None
        return next_base_class

    def get_all_test_methods(self) -> List[callable]:
        """
        This method returns all test methods that were defined in the related scenario. A testmethod has to start with
        `test_*`.
        """
        all_relevant_func = []

        all_methods = inspect.getmembers(self.related_cls, inspect.isfunction)
        for cur_method_name, cur_function in all_methods:
            if cur_method_name.startswith('test_'):
                all_relevant_func.append(cur_function)

        return all_relevant_func

    def validate_feature_clearance_for_parallel_connections(self):
        """
        This method validates for every active class-based feature (only the ones that have an active VDevice<->Device
        mapping), that there exist a clear scenario-device-connection for this feature. The method throws an
        :class:`UnclearAssignableFeatureConnectionError` if there exists more than one possible device-connection
        for the related devices and the method is not able to determine a clear connection.
        """

        for cur_from_device in self.get_all_abs_inner_device_classes():
            # determine all VDevice-Device mappings for this one, by iterating over all instantiated Feature classes
            cur_from_device_instantiated_features = \
                DeviceController.get_for(cur_from_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_from_device_instantiated_features.items():
                mapped_vdevice, mapped_device = cur_feature.active_vdevice_device_mapping
                if mapped_device is None:
                    # ignore this, because we have no vDevices here
                    continue

                # now check if one or more single of the classbased connection are CONTAINED IN the possible
                # parallel connection (only if there exists more than one parallel)
                feature_cnn = Connection.based_on(*FeatureController.get_for(
                        cur_feature.__class__).get_abs_class_based_for_vdevice()[mapped_vdevice])

                # search node names that is the relevant connection
                relevant_cnns: List[Connection] = []
                mapped_device_abs_cnns = DeviceController.get_for(mapped_device).get_all_absolute_connections()
                for _, all_connections in mapped_device_abs_cnns.items():
                    relevant_cnns += [cur_cnn for cur_cnn in all_connections
                                      if cur_cnn.has_connection_from_to(cur_from_device, mapped_device)]

                if len(relevant_cnns) <= 1:
                    # ignore if there are not more than one relevant connection
                    continue

                # there are some parallel connections -> check that only one fits with the feature
                matched_relevant_cnns = []
                for cur_relevant_cnn in relevant_cnns:
                    cur_relevant_cnn_singles = cur_relevant_cnn.get_singles()

                    for cur_relevant_single in cur_relevant_cnn_singles:
                        matches = [True for cur_feature_cnn in feature_cnn.get_singles()
                                   if cur_feature_cnn.contained_in(cur_relevant_single)]
                        if len(matches):
                            matched_relevant_cnns.append(True)
                            break
                if sum(matched_relevant_cnns) > 1:
                    raise UnclearAssignableFeatureConnectionError(
                        f"the devices {cur_from_device.__name__} and {mapped_device.__name__} have "
                        f"multiple parallel connections - the device `{cur_from_device.__name__}` uses a "
                        f"feature `{cur_feature.__class__.__name__}` that matches with the device "
                        f"`{mapped_device.__name__}`, but it is not clear which of the parallel connection "
                        f"could be used")

    def get_feature_cleaned_absolute_single_connections(self) -> \
            Tuple[Dict[Type[Device], Dict[str, Dict[Type[Device], Dict[str, List[Connection]]]]],  Dict[
                Tuple[Device, Device], List[Connection]]]:
        """
        This method returns all absolute-single connections between all devices of this scenario, but already cleaned
        based on the cumulated class-based decorators of all the feature devices.

        .. note::
            Please note, that the reduction candidates connections do not have to be unique.

        :return: returns a tuple with the cleaned up connections (sorted as dictionary per device) and the reduced
                 connections as second element
        """
        reduction_candidates = {}

        def add_reduction_candidate(device, other_device, connections: Connection):

            if (device, other_device) not in reduction_candidates.keys() and \
                    (other_device, device) not in reduction_candidates.keys():
                # we have to add it as new list
                reduction_candidates[(device, other_device)] = [connections]
            elif (device, other_device) in reduction_candidates.keys():
                reduction_candidates[(device, other_device)].append(connections)
            elif (other_device, device) in reduction_candidates.keys():
                reduction_candidates[(other_device, device)].append(connections)

        # start to generate the singles for every connection between the devices of every scenario
        all_abs_single_connections = self.get_absolute_single_connections()

        def reduce_based_on_feature_cnns_for_devices(
                feature_cnn: Connection, dev1: Type[Device], node_dev1: str, dev2: Type[Device], node_dev2: str):
            # execute further process only if there is exactly one relevant connection
            start_length_before_reduction = \
                len(all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2])
            for cur_abs_connection in \
                    all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2].copy():
                if not feature_cnn.contained_in(cur_abs_connection, ignore_metadata=True):
                    # this abs single connection is not fulfilled by the current feature -> remove it
                    all_abs_single_connections[dev1][node_dev1][
                        dev2][node_dev2].remove(cur_abs_connection)
                    add_reduction_candidate(dev1, dev2, cur_abs_connection)
            if start_length_before_reduction > 0 and \
                    len(all_abs_single_connections[dev1][node_dev1][
                            dev2][node_dev2]) == 0:
                raise ConnectionIntersectionError(
                    f"the `{self.related_cls.__name__}` has a connection from device "
                    f"`{dev1.__name__}` to `{dev2.__name__}` - some mapped VDevices of "
                    f"their feature classes define mismatched connections")

        def get_single_cnns_between_device_for_feature(from_device, to_device, relevant_feature_cnn):
            # search node names that is the relevant connection
            relevant_cnns: List[List[Connection]] = []
            for _, cur_node_data in all_abs_single_connections[from_device].items():
                for cur_to_device, cur_to_device_data in cur_node_data.items():
                    if cur_to_device == to_device:
                        relevant_cnns += [cur_cnns for _, cur_cnns in cur_to_device_data.items()]

            result_singles = None
            if len(relevant_cnns) > 1:
                # there exists parallel connections - filter only the relevant one
                for cur_single_cnns in relevant_cnns:
                    for cur_single_cnn in cur_single_cnns:
                        if relevant_feature_cnn.contained_in(cur_single_cnn):
                            # this is the relevant connection (all other can not fit, because we have
                            # already checked this with method
                            # `scenario_controller.validate_feature_clearance_for_parallel_connections()`)
                            result_singles = cur_single_cnns
                            break
                    if result_singles is not None:
                        break
            elif len(relevant_cnns) == 1:
                result_singles = relevant_cnns[0]
            if result_singles is None:
                raise ValueError("can not find relevant connection of all parallel connections")
            return result_singles

        all_devices = self.get_all_abs_inner_device_classes()
        for cur_from_device in all_devices:
            # determine all VDevice-Device mappings for this one, by iterating over all instantiated Feature classes
            cur_from_device_instantiated_features = \
                DeviceController.get_for(cur_from_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_from_device_instantiated_features.items():
                mapped_vdevice, mapped_device = cur_feature.active_vdevice_device_mapping
                if mapped_device is None:
                    # ignore this, because we have no vDevices here
                    continue

                # now try to reduce the scenario connections according to the requirements of the feature class
                cur_feature_class_based_for_vdevice = \
                    FeatureController.get_for(
                        cur_feature.__class__).get_abs_class_based_for_vdevice()[mapped_vdevice]
                cur_feature_cnn = Connection.based_on(*cur_feature_class_based_for_vdevice)

                device_cnn_singles = get_single_cnns_between_device_for_feature(
                    from_device=cur_from_device, to_device=mapped_device, relevant_feature_cnn=cur_feature_cnn)

                reduce_based_on_feature_cnns_for_devices(
                    cur_feature_cnn, device_cnn_singles[0].from_device, device_cnn_singles[0].from_node_name,
                    device_cnn_singles[0].to_device, device_cnn_singles[0].to_node_name)
                # do the same for the opposite direction (features are always bidirectional)
                reduce_based_on_feature_cnns_for_devices(
                    cur_feature_cnn, device_cnn_singles[0].to_device, device_cnn_singles[0].to_node_name,
                    device_cnn_singles[0].from_device, device_cnn_singles[0].from_node_name)

        return all_abs_single_connections, reduction_candidates

    def determine_absolute_device_connections(self):
        """
        This method determines the real possible Sub-Connections for every element of the scenarios. For this the method
        will create a possible intersection connection, for the :class:´Connection´ between two devices and
        all :class:`Connection`-Subtrees that are allowed for the mapped vDevices in the used :class:`Feature`
        classes.
        The data will be saved in the :class:`Device` property ``_absolute_connections``. If the method detects an empty
        intersection between two devices that are connected through a VDevice-Device mapping, the method will throw an
        exception.
        """

        # start to generate the singles for every connection between the devices of every scenario
        all_abs_single_connections, reduction_candidates = self.get_feature_cleaned_absolute_single_connections()

        # generate all required warnings
        for cur_warning_tuple in reduction_candidates:
            logger.warning(f"detect some connections between the devices `{cur_warning_tuple[0].__name__}` and "
                           f"`{cur_warning_tuple[1].__name__}` of scenario `{self.related_cls.__name__}` that can be "
                           f"reduced, because their related features only use a subset of the defined connection")

        # first cleanup the relevant absolute connections
        for cur_from_device, from_device_data in all_abs_single_connections.items():
            for _, from_node_data in from_device_data.items():
                for cur_to_device, to_device_data in from_node_data.items():
                    cur_from_device_controller = DeviceController.get_for(cur_from_device)
                    cur_to_device_controller = DeviceController.get_for(cur_to_device)

                    cur_from_device_controller.cleanup_absolute_connections_with(cur_to_device)
                    cur_to_device_controller.cleanup_absolute_connections_with(cur_from_device)

        # replace all absolute connection with the single ones
        for cur_from_device, from_device_data in all_abs_single_connections.items():
            for _, from_node_data in from_device_data.items():
                for cur_to_device, to_device_data in from_node_data.items():
                    for _, cur_single_cnns in to_device_data.items():

                        cur_from_device_controller = DeviceController.get_for(cur_from_device)
                        cur_to_device_controller = DeviceController.get_for(cur_to_device)

                        new_cnn = Connection.based_on(*cur_single_cnns)
                        new_cnn.set_metadata_for_all_subitems(cur_single_cnns[0].metadata)
                        if cur_from_device == cur_single_cnns[0].from_device:
                            cur_from_device_controller.add_new_absolute_connection(new_cnn)
                        else:
                            cur_to_device_controller.add_new_absolute_connection(new_cnn)



import balder
from ...lib.features import FeatureII
from ...lib.connections import ChildBConnection
from ...balderglob import RuntimeObserver
from .feature_i import SetupFeatureI


# this forces that the mapped vDevice has to have a `ChildBConnection` - all other of the conn-tree has to be removed
# for further processing
@balder.for_vdevice("VDeviceFeatureI", ChildBConnection)
class SetupFeatureII(FeatureII):

    class VDeviceFeatureI(balder.VDevice):
        vdevice_i = SetupFeatureI()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.do_something, "enter `SetupFeatureII.do_something`",
            category="feature")

    def do_something_special_with_the_vdevice(self):
        super(SetupFeatureII, self).do_something_special_with_the_vdevice()
        self.VDeviceFeatureI.vdevice_i.do_something_special()



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0NoMultipleVariationsForWrongConnection(Base0EnvtesterClass):
    """
    This testcase uses an expanded version of the envtester environment. The ``SetupA`` has two setup devices that
    basically implements the same features as the ``ScenarioDevice2`` in the ``ScenarioA``. The ``SetupDevice2b``
    has a Connection that does not fulfill the requirement of the matched scenario. We expect that balder
    will create only one variation for the ``SetupDevice2a`` within the ``ScenarioA`` <=> ``SetupA`` constellation.

    This should not influence the single mapping for ``ScenarioB`` <=> ``SetupB``.
    """

    @property
    def expected_data(self) -> tuple:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            [
                (
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                             "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_1
                            {"cls": "ScenarioA", "meth": "test_a_1"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase",
                             "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_2
                            {"cls": "ScenarioA", "meth": "test_a_2"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_1
                            {"cls": "ScenarioB", "meth": "test_b_1"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_2
                            {"cls": "ScenarioB", "meth": "test_b_2"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
            ],
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.SUCCESS, \
            "test session does not terminates with success"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.SUCCESS, \
            "global executor tree body part does not set ResultState.SUCCESS"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                "the setup executor does not have result SUCCESS"

            assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
            assert cur_setup_executor.body_result.result == ResultState.SUCCESS
            assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result SUCCESS"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result SUCCESS"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from __future__ import annotations
from typing import Type, Union, List, TYPE_CHECKING

from _balder.testresult import ResultState, BranchBodyResult
from _balder.executor.basic_executor import BasicExecutor
from _balder.executor.scenario_executor import ScenarioExecutor
from _balder.previous_executor_mark import PreviousExecutorMark

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.scenario import Scenario
    from _balder.fixture_manager import FixtureManager
    from _balder.executor.executor_tree import ExecutorTree


class SetupExecutor(BasicExecutor):
    """
    A SetupExecutor is the highest branch object of an :class:`ExecutorTree`. It contains all scenarios and the
    underlying device mappings and their test cases that exist with this setup.
    """

    def __init__(self, setup: Type[Setup], parent: ExecutorTree):
        super().__init__()
        self._scenario_executors: List[ScenarioExecutor] = []
        # check if instance already exists
        if hasattr(setup, "_instance") and setup._instance is not None and isinstance(setup._instance, setup):
            self._base_setup_class = setup._instance
        else:
            self._base_setup_class = setup()
            setup._instance = self._base_setup_class
        self._parent_executor = parent
        self._fixture_manager = parent.fixture_manager

        # contains the result object for the BODY part of this branch
        self.body_result = BranchBodyResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> List[ScenarioExecutor]:
        return self.scenario_executors

    @property
    def parent_executor(self) -> ExecutorTree:
        return self._parent_executor

    @property
    def base_instance(self) -> object:
        """returns the base class instance to which this executor instance belongs"""
        return self.base_setup_class

    @property
    def base_setup_class(self) -> Setup:
        """returns the base :class:`Setup` that belongs to this executor"""
        return self._base_setup_class

    @property
    def scenario_executors(self) -> List[ScenarioExecutor]:
        """returns a list with all scenario executors that belongs to this setup executor"""
        return self._scenario_executors

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the current active fixture manager for this executor"""
        return self._fixture_manager

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self):
        print(f"SETUP {self.base_setup_class.__class__.__name__}")

    def _body_execution(self):
        for cur_scenario_executor in self.scenario_executors:
            if cur_scenario_executor.has_runnable_elements():
                cur_scenario_executor.execute()
            elif cur_scenario_executor.prev_mark == PreviousExecutorMark.SKIP:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.SKIP)
            elif cur_scenario_executor.prev_mark == PreviousExecutorMark.COVERED_BY:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.COVERED_BY)
            else:
                cur_scenario_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self):
        pass

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def cleanup_empty_executor_branches(self):
        to_remove_executor = []
        for cur_scenario_executor in self.scenario_executors:
            cur_scenario_executor.cleanup_empty_executor_branches()
            if len(cur_scenario_executor.variation_executors) == 0:
                # remove this whole executor because it has no children anymore
                to_remove_executor.append(cur_scenario_executor)
        for cur_scenario_executor in to_remove_executor:
            self._scenario_executors.remove(cur_scenario_executor)

    def add_scenario_executor(self, scenario_executor: ScenarioExecutor):
        """
        This method adds a new ScenarioExecutor to the child element list of the tree
        """
        if not isinstance(scenario_executor, ScenarioExecutor):
            raise TypeError("the given object `scenario_executor` must be of type type `ScenarioExecutor`")
        if scenario_executor in self._scenario_executors:
            raise ValueError("the given object `scenario_executor` already exists in child list")
        self._scenario_executors.append(scenario_executor)

    def get_executor_for_scenario(self, scenario: Type[Scenario]) -> Union[ScenarioExecutor, None]:
        """
        This method searches for a ScenarioExecutor in the internal list for which the given scenario is
        contained in

        :param scenario: the scenario class for which the executor should be searched for

        :return: returns the associated ScenarioExecutor or None if no matching could be found
        """
        for cur_scenario_executor in self._scenario_executors:
            if cur_scenario_executor.base_scenario_class.__class__ == scenario:
                return cur_scenario_executor
        # can not find some
        return None



from _balder.balder_session import BalderSession
from _balder.exceptions import VDeviceOverwritingError, BalderException

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0OverwriteIllegallyMissInheritBefsce(Base0EnvtesterClass):
    """
    This testcase is a modified version of the basic ENVTESTER environment. It is limited to the `ScenarioA` and the
    `SetupA` and uses a feature `FeatureOfRelevance` for testing the correct overwriting of VDevices BEFORE the feature
    is being instantiated in a SCENARIO.

    This test defines a new feature `FeatureOfRelevanceLvl1` with two vDevices. This feature will be overwritten by the
    direct child `FeatureOfRelevanceLvl2`, that will be used in our scenario and also be mapped to a vDevice. Instead,
    of made as in the test `test_0_overwrite_correctly_befsce`, the inheritance is missing for one vDevices here.
    Of course the setup device also implements a child class (`SetupFeatureOfRelevanceLvl3`) of it.

    The test should fail on collector level, because overwritten vDevices with the same name have to inherit from
    itself too.
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self) -> int:
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, VDeviceOverwritingError), 'unexpected exception type'
        assert exc.args[0] == "the inner vDevice class `FeatureOfRelevanceLvl2.VDeviceWithI` has the same name than " \
                              "the vDevice `FeatureOfRelevanceLvl1.VDeviceWithI` - it should also inherit from it"

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree is None, "test session does not terminates before collector work was done"
